package de.moritzruth.lampenfieber.device

import de.moritzruth.theaterdsl.device.AngleDV
import de.moritzruth.theaterdsl.device.ConcreteDV
import de.moritzruth.theaterdsl.device.Device
import de.moritzruth.theaterdsl.device.PercentageDV
import de.moritzruth.theaterdsl.dmx.DmxAddress
import de.moritzruth.theaterdsl.dmx.DmxDataWriter
import de.moritzruth.theaterdsl.dmx.DmxValue
import de.moritzruth.theaterdsl.value.Angle
import de.moritzruth.theaterdsl.value.Percentage
import de.moritzruth.theaterdsl.value.degrees
import kotlinx.collections.immutable.persistentSetOf
import kotlin.math.roundToInt
import kotlin.time.Duration.Companion.minutes

class FuturelightDmh160(override val firstChannel: DmxAddress, private val isRight: Boolean) : Device {
    companion object {
        private const val PAN_LIMIT = 630.0 // degrees
        private const val TILT_LIMIT = 270.0 // degrees
    }

    override val numberOfChannels: UInt = 16u

    override fun writeDmxData(writer: DmxDataWriter, isLightBehindCurtainOn: Boolean) {
        val pan = (pan.getCurrentValue().degrees.mod(PAN_LIMIT) / PAN_LIMIT * 65536).roundToInt().toUShort()
        val tilt = (tilt.getCurrentValue().degrees.mod(TILT_LIMIT) / TILT_LIMIT * 65536).roundToInt().toUShort()

        writer.writeHighByte(pan)
        writer.writeLowByte(pan)
        writer.writeHighByte(tilt)
        writer.writeLowByte(tilt)
        writer.writeRaw(DmxValue(20u)) // pan/tilt speed = idk, 255 is slowest, but 0 is also slow

        val strobeSpeed = strobeSpeed.getCurrentValue()
        writer.writeRaw(
            if (strobeSpeed.value == 0.0) DmxValue(255u)
            else DmxValue((strobeSpeed.value * 31 + 64).roundToInt().toUByte())
        )

        writer.writePercentage(brightness.getCurrentValue())
        writer.writeRaw(colorWheelMode.getCurrentValue().getDmxValue())
        writer.writeRaw(DmxValue(0u)) // gobo wheel = open
        writer.writeRaw(DmxValue(0u)) // gobo rotation = none
        writer.writeRaw(DmxValue(0u)) // second gobo = none
        writer.writeRaw(prismMode.getCurrentValue().dmxValue)
        writer.writeRaw(DmxValue((prismRotationSpeed.getCurrentValue().value * 123 + 4).roundToInt().toUByte()))
        writer.writeRaw(DmxValue(255u)) // focus = maximum distance
        writer.writeRaw(DmxValue(0u)) // iris = ?
        writer.writeRaw(DmxValue(0u)) // functions = noop
    }

    sealed interface ColorWheelMode {
        fun getDmxValue(): DmxValue
        sealed class Simple(staticDmxValue: Int) : ColorWheelMode {
            val staticDmxValue = DmxValue(staticDmxValue.toUByte())
            override fun getDmxValue() = staticDmxValue
        }

        class Rotate(val speed: Percentage, val backwards: Boolean = false) : ColorWheelMode {
            override fun getDmxValue() = run {
                if (backwards) (speed.value * 15).roundToInt() + 240
                else (speed.value * 15).roundToInt() + 224
            }.let { DmxValue(it.toUByte()) }
        }

        object White : Simple(0)
        object DarkPink : Simple(49)
        object DarkRed : Simple(7)
        object Orange : Simple(43)
        object ToxicGreen : Simple(19)
        object LightBlue : Simple(37)
        object DarkBlue : Simple(67)
        object LightPink : Simple(31)
        object Yellow : Simple(25)
    }

    enum class PrismMode(val dmxValue: DmxValue) {
        OPEN(DmxValue(0u)),
        FACETS_3(DmxValue(64u)),
        FACETS_8(DmxValue(128u)),
        FROST(DmxValue(192u))
    }

    val pan = AngleDV(maximumValue = Angle(PAN_LIMIT))
    val tilt = AngleDV()
    val brightness = PercentageDV()
    val colorWheelMode = ConcreteDV<ColorWheelMode>(ColorWheelMode.White)
    val strobeSpeed = PercentageDV()
    val prismMode = ConcreteDV(PrismMode.OPEN)
    val prismRotationSpeed = PercentageDV()

    fun startRoomMovement(rotationsPerMinute: Double) { // TODO: Test
        val period = (1.minutes / rotationsPerMinute) * 2
        if (isRight) pan.rampUp(period * (PAN_LIMIT / 360.0))
        else pan.rampDown(period * (PAN_LIMIT / 360.0))
        tilt.sine(period, minimum = 50.degrees, maximum = 90.degrees, start = 50.degrees)
    }

    override val dvs = persistentSetOf(
        pan, tilt, brightness, colorWheelMode, strobeSpeed
    )

    fun pointAtStageMiddleSingleCenterMarking() {
        if (isRight) {
            tilt.static(209.degrees)
            pan.static(26.degrees)
        } else {
            tilt.static(203.degrees)
            pan.static(400.degrees)
        }
    }

    fun pointAtStageMiddleOffCenterMarking() {
        if (isRight) {
            tilt.static(205.degrees)
            pan.static(15.degrees)
        } else {
            tilt.static(201.degrees)
            pan.static(410.degrees)
        }
    }
}