package de.moritzruth.theaterdsl.show

import de.moritzruth.theaterdsl.runOptionalSanityChecks
import de.moritzruth.theaterdsl.util.StringWithDetails
import io.github.oshai.KotlinLogging
import kotlinx.collections.immutable.toImmutableList
import kotlinx.collections.immutable.toImmutableMap
import kotlinx.collections.immutable.toImmutableSet

@DslMarker
annotation class TheaterDslMarker

@TheaterDslMarker
interface ShowBuilderContext {
    fun act(name: String, build: ActBuilderContext.() -> Unit)
}

@TheaterDslMarker
interface ActBuilderContext {
    fun scene(name: String, build: SceneBuilderContext.() -> Unit)
}

@TheaterDslMarker
interface SceneBuilderContext {
    fun step(cue: StepCue, build: StepDataBuilderContext.() -> Unit)
    fun lightStep(cue: StepCue, runner: StepRunner)
}

@TheaterDslMarker
interface StepDataBuilderContext {
    val props: PropsBuilderMap
    var leftSpotTarget: String?
    var rightSpotTarget: String?
    var curtainState: CurtainState

    fun actors(build: ActorsBuildContext.() -> Unit)
    fun onRun(block: StepRunner)
}

class PropsBuilderMap(private val backingMap: MutableMap<PropPosition, StringWithDetails?>) {
    operator fun set(position: PropPosition, prop: String?) = backingMap.set(position, prop?.let { StringWithDetails(it) })

    operator fun invoke(block: (PropsBuilderMap) -> Unit) {
        block(this)
    }
}

@TheaterDslMarker
class ActorsBuildContext(
    private val entrances: MutableSet<StringWithDetails>,
    private val exits: MutableSet<StringWithDetails>
) {
    operator fun StringWithDetails.unaryPlus() {
        entrances.add(this)
    }

    operator fun String.unaryPlus() {
        entrances.add(StringWithDetails(this))
    }

    operator fun StringWithDetails.unaryMinus() {
        exits.add(this)
    }

    operator fun String.unaryMinus() {
        exits.add(StringWithDetails(this))
    }
}

@TheaterDslMarker
fun createShow(build: ShowBuilderContext.() -> Unit): Show {
    val acts = mutableListOf<Act>()

    object : ShowBuilderContext {
        override fun act(name: String, build: ActBuilderContext.() -> Unit) {
            acts.add(buildAct(acts.size, name, build))
        }
    }.build()

    return Show(acts.toList())
}

@TheaterDslMarker
private fun buildAct(actIndex: Int, actName: String, build: ActBuilderContext.() -> Unit): Act {
    val scenes = mutableListOf<Scene>()
    val actorsOnStage = mutableListOf<String>()
    val props = PropPosition.values().associateWith<PropPosition, StringWithDetails?> { null }.toMutableMap()

    var leftSpotTarget_: String? = null
    var rightSpotTarget_: String? = null
    var curtainState_ = CurtainState.CLOSED

    object : ActBuilderContext {
        override fun scene(@Suppress("PARAMETER_NAME_CHANGED_ON_OVERRIDE") sceneName: String, build: SceneBuilderContext.() -> Unit) {
            val steps = mutableListOf<Step>()

            object : SceneBuilderContext {
                override fun lightStep(cue: StepCue, runner: StepRunner) {
                    step(cue) {
                        onRun(runner)
                    }
                }

                override fun step(cue: StepCue, build: StepDataBuilderContext.() -> Unit) {
                    val changedProps = mutableMapOf<PropPosition, StringWithDetails?>()
                    val actorEntrances = mutableSetOf<StringWithDetails>()
                    val actorExits = mutableSetOf<StringWithDetails>()
                    var runner: StepRunner? = null

                    object : StepDataBuilderContext {
                        override val props = PropsBuilderMap(changedProps)

                        override var rightSpotTarget: String?
                            get() = rightSpotTarget_
                            set(value) {
                                rightSpotTarget_ = value
                            }

                        override var leftSpotTarget: String?
                            get() = leftSpotTarget_
                            set(value) {
                                leftSpotTarget_ = value
                            }

                        override var curtainState: CurtainState
                            get() = curtainState_
                            set(value) {
                                curtainState_ = value
                            }

                        override fun actors(build: ActorsBuildContext.() -> Unit) {
                            ActorsBuildContext(actorEntrances, actorExits).build()
                        }

                        override fun onRun(block: StepRunner) {
                            runner = block
                        }
                    }.build()

                    val logger = KotlinLogging.logger("createAct / $actName / $sceneName / #${steps.size + 1} ${cue.format()}")

                    val actorEntrancesNames = actorEntrances.map { it.main }
                    val actorExitsNames = actorExits.map { it.main }

                    runOptionalSanityChecks {
                        val actorEntrancesDuplicateNames = actorEntrancesNames.groupingBy { it }.eachCount().filter { it.value > 1 }.keys
                        if (actorEntrancesDuplicateNames.isNotEmpty())
                            logger.warn("Duplicate actor entrances: ${actorEntrancesDuplicateNames.joinToString()}")

                        val actorExitsDuplicateNames = actorExitsNames.groupingBy { it }.eachCount().filter { it.value > 1 }.keys
                        if (actorExitsDuplicateNames.isNotEmpty())
                            logger.warn("Duplicate actor exits: ${actorExitsDuplicateNames.joinToString()}")

                        val removedAndAddedActorNames = actorEntrancesNames.intersect(actorExitsNames.toSet())
                        if (removedAndAddedActorNames.isNotEmpty()) logger.warn(
                            "These actors are removed and added in the same step: " +
                                    removedAndAddedActorNames.joinToString()
                        )

                        val removedActorsNotOnStage = actorExitsNames.subtract(actorsOnStage.toSet())
                        if (removedActorsNotOnStage.isNotEmpty())
                            logger.warn("These actors cannot exit because they are not on the stage: ${removedActorsNotOnStage.joinToString()}")

                        val addedActorsAlreadyOnStage = actorsOnStage.intersect(actorEntrancesNames.toSet())
                        if (addedActorsAlreadyOnStage.isNotEmpty())
                            logger.warn("These actors cannot enter because they are already on the stage: ${addedActorsAlreadyOnStage.joinToString()}")
                    }

                    actorsOnStage.removeAll(actorExitsNames.toSet())
                    actorsOnStage.addAll(actorEntrancesNames)

                    changedProps.forEach { (k, v) -> props[k] = v }

                    steps.add(
                        Step(
                            ShowPosition(actIndex, scenes.size, steps.size),
                            cue,
                            actorEntrances.toImmutableSet(),
                            actorExits.toImmutableSet(),
                            actorsOnStage.toImmutableList(),
                            props.toImmutableMap(),
                            changedProps.isNotEmpty(),
                            leftSpotTarget_,
                            rightSpotTarget_,
                            curtainState_,
                            runner
                        )
                    )
                }
            }.build()

            scenes.add(Scene(sceneName, steps.toImmutableList()))
        }
    }.build()

    return Act(actName, scenes.toImmutableList())
}