package de.moritzruth.theaterdsl.show

import de.moritzruth.lampenfieber.device.fogMachine
import de.moritzruth.theaterdsl.device.Device
import de.moritzruth.theaterdsl.device.DynamicValue
import de.moritzruth.theaterdsl.dmx.EnttecOpenDmxUsb
import de.moritzruth.theaterdsl.dmx.PerDeviceDmxDataWriter
import de.moritzruth.theaterdsl.util.InstantAsEpochMillisecondsSerializer
import de.moritzruth.theaterdsl.util.mapState
import de.moritzruth.theaterdsl.value.Percentage
import de.moritzruth.theaterdsl.value.percent
import io.github.oshai.KLogger
import io.github.oshai.KotlinLogging
import io.ktor.http.*
import io.ktor.serialization.kotlinx.json.*
import io.ktor.server.application.*
import io.ktor.server.cio.*
import io.ktor.server.engine.*
import io.ktor.server.http.content.*
import io.ktor.server.plugins.contentnegotiation.*
import io.ktor.server.request.*
import io.ktor.server.response.*
import io.ktor.server.routing.*
import io.ktor.server.websocket.*
import io.ktor.websocket.*
import kotlinx.collections.immutable.ImmutableSet
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.*
import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import java.util.concurrent.atomic.AtomicInteger
import kotlin.math.roundToLong
import kotlin.system.measureTimeMillis
import kotlin.time.Duration.Companion.seconds
import kotlin.time.TimeSource
import kotlin.time.toJavaDuration

data class FogState(val power: Percentage, val time: TimeSource.Monotonic.ValueTimeMark = TimeSource.Monotonic.markNow()) {
    companion object {
        fun off() = FogState(0.percent)
    }
}

class ShowContext(
    val devices: ImmutableSet<Device>,
    val show: Show,
    val logger: KLogger,
    val stateFlow: MutableStateFlow<ShowState>,
    val fogState: MutableStateFlow<FogState>,
    val outputDataFreeze: AtomicInteger
) {
    val stepFlow = stateFlow.mapState {
        if (it.position == ShowPosition.START) Step.START
        else show.acts[it.position.act].scenes[it.position.scene].steps[it.position.step]
    }

    fun goTo(position: ShowPosition) {
        val step = show.acts[position]

        stateFlow.update {
            val isLightBehindCurtainOn = if (step.curtainState == CurtainState.OPEN) false else it.isLightBehindCurtainOn

            when {
                step.cue is StepCue.MusicStart -> {
                    it.copy(
                        position = position,
                        activeMusic = ActiveShowMusic(
                            title = step.cue.title,
                            duration = step.cue.duration
                        ),
                        musicStartTime = Clock.System.now(),
                        isLightBehindCurtainOn = isLightBehindCurtainOn
                    )
                }

                step.cue is StepCue.MusicEnd ||
                        step.position == ShowPosition.START -> it.copy(position = position, activeMusic = null, isLightBehindCurtainOn = isLightBehindCurtainOn)

                else -> it.copy(position = position, isLightBehindCurtainOn = isLightBehindCurtainOn)
            }
        }

        if (step.cue is StepCue.MusicStart) {
            logger.info("Music started: ${step.cue.title} (${step.cue.duration.inWholeSeconds}s)")
        }
    }
}

suspend fun runShow(show: Show, devices: ImmutableSet<Device>) = coroutineScope {
    val logger = KotlinLogging.logger("runShow")

    val stateFlow = MutableStateFlow(
        ShowState(
            ShowPosition.START,
            "",
            null,
            Instant.DISTANT_PAST,
            false
        )
    )

    val context = ShowContext(
        devices,
        show,
        logger,
        stateFlow,
        MutableStateFlow(FogState.off()),
        AtomicInteger(0)
    )

    startDataWriting(context)
    startStepRunning(context)
    startFogHandling(context)
    startWebsocketServer(context)

    launch {
        stateFlow.collect {
            logger.info("ShowState changed: $it")
        }
    }
}

fun CoroutineScope.startDataWriting(context: ShowContext) = launch {
    while (isActive) {
        val took = measureTimeMillis {
            val data = UByteArray(512)
            val writer = PerDeviceDmxDataWriter(data)

            for (device in context.devices) {
                writer.reset(0, device.firstChannel, device.numberOfChannels)
                device.writeDmxData(writer, context.stateFlow.value.isLightBehindCurtainOn)
            }

            if (context.outputDataFreeze.get() == 0) EnttecOpenDmxUsb.data = data
        }

        delay(((1000.0 / 30) - took).roundToLong())
    }
}

fun CoroutineScope.startFogHandling(context: ShowContext) = launch {
    context.fogState.collectLatest { state ->
        fogMachine.power.static(state.power)

        delay(1000)
        fogMachine.power.off()
    }
}

fun CoroutineScope.startStepRunning(context: ShowContext) = launch {
    var lastPosition = ShowPosition.START
    var lastStepJob: Job? = null

    context.stateFlow.map { it.position }.distinctUntilChanged().collect { position ->
        context.outputDataFreeze.incrementAndGet()

        if (position < lastPosition) {
            context.devices.forEach { it.dvs.forEach(DynamicValue<*>::reset) }
            lastPosition = ShowPosition.START
        }

        while (lastPosition != position) {
            lastPosition = context.show.getNextValidPosition(lastPosition) ?: break

            val step = context.show.acts[lastPosition]
            lastStepJob?.cancelAndJoin()
            lastStepJob = launch(SupervisorJob(currentCoroutineContext().job), CoroutineStart.UNDISPATCHED) {
                val runContext = object : StepRunContext, CoroutineScope by this {}

                step.runner?.let { runContext.it() }
            }
        }

        context.outputDataFreeze.decrementAndGet()
    }
}

@Serializable
data class StateUpdateMessage(
    @Serializable(with = InstantAsEpochMillisecondsSerializer::class)
    val timestamp: Instant,
    val state: ShowState
)

private fun CoroutineScope.startWebsocketServer(context: ShowContext) = launch(Dispatchers.IO) {
    val updateMessageJsonFlow = context.stateFlow.map { state ->
        Json.encodeToString(
            StateUpdateMessage(
                Clock.System.now(),
                state
            )
        )
    }

    val showJson = Json.encodeToString(context.show)
    val indexHtmlContent = this::class.java.getResourceAsStream("/ui/index.html")!!.reader().use { it.readText() }

    embeddedServer(CIO, port = 8000) {
        install(WebSockets) {
            pingPeriod = 10.seconds.toJavaDuration()
            timeout = 15.seconds.toJavaDuration()
            maxFrameSize = Long.MAX_VALUE
        }

        install(ContentNegotiation) {
            json()
        }

        routing {
            route("/api") {
                post("/go") {
                    val position = call.receive<ShowPosition>()
                    context.goTo(position)
                    call.respond(HttpStatusCode.NoContent)
                }

                post("/message") {
                    val message = call.receiveText()
                    context.stateFlow.update { it.copy(message = message) }
                    call.respond(HttpStatusCode.NoContent)
                }

                post("/fog") {
                    val power = call.receive<Percentage>()
                    context.fogState.value = FogState(power)
                    call.respond(HttpStatusCode.NoContent)
                }

                post("/lightBehindCurtain") {
                    val isOn = call.receive<Boolean>()
                    if (context.stepFlow.value.curtainState == CurtainState.CLOSED || !isOn) {
                        context.stateFlow.update { it.copy(isLightBehindCurtainOn = isOn) }
                    }

                    call.respond(HttpStatusCode.NoContent)
                }

                get("/show") {
                    call.respond(context.show.acts.toMutableList())
                }

                get("/state") {
                    call.respond(context.stateFlow.value)
                }

                val clientsCount = AtomicInteger(0)

                webSocket("/ws") {
                    var count = clientsCount.incrementAndGet()
                    context.logger.info("[${call.request.local.remoteHost}] connected ($count)")

                    val sendJob = launch {
                        send(Frame.Text(showJson))
                        updateMessageJsonFlow.collect { send(Frame.Text(it)) }
                    }

                    try {
                        for (frame in incoming) {
                            context.logger.debug { "[${call.request.local.remoteHost}] frame: $frame" }
                        }
                    } finally {
                        sendJob.cancel()
                        count = clientsCount.decrementAndGet()
                        context.logger.info("[${call.request.local.remoteHost}] disconnected ($count)")
                    }
                }
            }

            staticResources("/assets", "ui/assets") {
                enableAutoHeadResponse()
            }

            get("/{...}") {
                call.respondText(indexHtmlContent, ContentType.Text.Html, HttpStatusCode.OK)
            }
        }
    }.start(wait = true)
}